/* $NetBSD: $ */

/*-
  * Copyright (c) 2013 UPMC/LIP6
  * All rights reserved.
  * This software is distributed under the following condiions
  * compliant with the NetBSD foundation policy.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  */

/* driver for the VHDL PS/2 keyboard controller */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: machdep.c,v 1.223 2008/07/02 17:28:56 ad Exp $");

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/conf.h>
#include <sys/kauth.h>
#include <sys/kthread.h>
#include <sys/condvar.h>

#include <dev/pckbport/pckbportvar.h>

#include <machine/autoconf.h>
#include <machine/bus.h>

#undef KBC_DEBUG
#ifdef KBC_DEBUG
#define DPRINTF(m, x) {if (vcidebug_mask & (m)) printf x;}
#define KBCDEBUG_FUNC	0x1
#define KBCDEBUG_XFER	0x2
#define KBCDEBUG_XFER2	0x4
#define KBCDEBUG_INIT	0x8
static int vcidebug_mask = KBCDEBUG_XFER;
#else
#define DPRINTF(m, x)
#endif

/* KBC device registers */
#define KBC_DATA0		0x00
#define KBC_STATUS		0x01

#define KBC_STATUS_KBC0		(1 << 0)

#define KBC_DEVICE_REG_READ(sc, reg) \
	bus_space_read_4((sc)->sc_tag, (sc)->sc_handle, (reg) * 4)
#define KBC_DEVICE_REG_WRITE(sc, reg, val) \
	bus_space_write_4((sc)->sc_tag, (sc)->sc_handle, (reg) * 4, (val))

static int  vcikbc_match(device_t, cfdata_t, void *);
static void vcikbc_attach(device_t, device_t, void *);
static void vcikbc_io_thread(void *);

struct vcikbc_softc {
	device_t sc_dev;
	bus_space_tag_t sc_tag;
	bus_space_handle_t sc_handle;
	irq_t sc_ih;
	kmutex_t sc_mtx;
	lwp_t	   *sc_lwp;
	struct pckbport_tag *sc_pt;
	int sc_polling;
};

static int vcikbc_xt_translation(void *, pckbport_slot_t, int);
static int vcikbc_send_devcmd(void *, pckbport_slot_t, u_char);
static int vcikbc_poll_data1(void *, pckbport_slot_t);
static void vcikbc_slot_enable(void *, pckbport_slot_t, int);
static void vcikbc_intr_establish(void *, pckbport_slot_t);
static void vcikbc_set_poll(void *, pckbport_slot_t, int);

static struct pckbport_accessops const vcikbc_ops = {
        vcikbc_xt_translation,
	vcikbc_send_devcmd,
	vcikbc_poll_data1,
	vcikbc_slot_enable,
	vcikbc_intr_establish,
	vcikbc_set_poll
};

CFATTACH_DECL_NEW(vcikbc, sizeof(struct vcikbc_softc),
    vcikbc_match, vcikbc_attach, NULL, NULL);

int
vcikbc_match(device_t parent, cfdata_t match, void *aux)
{
	struct tsardevs_attach_args *tsd = aux;
	if (strcmp(tsd->tsd_devtype, "vci:ps2") != 0)
		return 0;
	return 1;
}

void
vcikbc_attach(device_t parent, device_t self, void *aux)
{
	struct vcikbc_softc *sc = device_private(self);
	struct tsardevs_attach_args *tsd = aux;

	aprint_normal(": PS/2 controller\n");

	sc->sc_dev = self;
	sc->sc_tag = tsd->tsd_tag;

	mutex_init(&sc->sc_mtx, MUTEX_DEFAULT, IPL_TTY);

	/* map the device */
	if (bus_space_map(sc->sc_tag,
	    tsd->tsd_reg[0].reg_addr,tsd->tsd_reg[0].reg_size, 0,
	    &sc->sc_handle)) {
			aprint_error_dev(self, "couldn't map registers\n");
			return;
	}
#ifdef notyet
	sc->sc_ih = intr_establish(tsd->tsd_xicu,
	    IRQ_HWI, tsd->tsd_irq, IPL_BIO,
	    device_xname(self), buf, vcikbc_intr, sc, NULL);
	if (sc->sc_ih == NULL) {
		aprint_error_dev(self, "can't establish interrupt\n");
		return;
	} else {
		aprint_normal_dev(self, "interrupting at %s\n", buf);
	}
#endif
	sc->sc_polling = 1;
	sc->sc_pt = pckbport_attach(sc, &vcikbc_ops);
	pckbport_attach_slot(sc->sc_dev, sc->sc_pt, PCKBPORT_KBD_SLOT);

	if (kthread_create(PRI_BIO, KTHREAD_MPSAFE, NULL,
	    vcikbc_io_thread, sc, &sc->sc_lwp, "%s", device_xname(self))) {
		aprint_error_dev(self, "couldn't create task thread\n");
		return;
	}
	return;
}

static void
vcikbc_io_thread(void *arg)
{
        struct vcikbc_softc *sc = (struct vcikbc_softc *)arg;
	uint32_t data;

	for(;;) {
		kpause("vcikbc", 0, 1, NULL);
		if (sc->sc_polling == 1)
			continue;

		if ((KBC_DEVICE_REG_READ(sc, KBC_STATUS) & KBC_STATUS_KBC0)
		    == 0)
			continue;
		mutex_enter(&sc->sc_mtx);
		if ((KBC_DEVICE_REG_READ(sc, KBC_STATUS) & KBC_STATUS_KBC0)
		    == 0) {
			mutex_exit(&sc->sc_mtx);
			continue;
		}
		data = KBC_DEVICE_REG_READ(sc, KBC_DATA0);
		DPRINTF(KBCDEBUG_XFER, ("pckbc read 0x%x\n", data));
		pckbportintr(sc->sc_pt, PCKBPORT_KBD_SLOT, data);
		mutex_exit(&sc->sc_mtx);
	}
}

static int
vcikbc_send_devcmd(void *arg, pckbport_slot_t slot, u_char cmd)
{
	DPRINTF(KBCDEBUG_XFER2, ("vcikbc_send_devcmd slot %d cmd %d\n"
	    slot, cmd));
	return 1;
}

static int
vcikbc_poll_data1(void *arg, pckbport_slot_t slot)
{
	struct vcikbc_softc *sc = (struct vcikbc_softc *)arg;
	int timeout;

	timeout = 10000;
	while ((KBC_DEVICE_REG_READ(sc, KBC_STATUS) & KBC_STATUS_KBC0) == 0) {
		DELAY(10);
		if (--timeout == 0) return -1;
	}
	return KBC_DEVICE_REG_READ(sc, KBC_DATA0);
}

static int
vcikbc_xt_translation(void *arg, pckbport_slot_t slot, int on)
{
	if (on)
		return 0; /* can't do xt translation */
	else
		return 1;
}

static void
vcikbc_slot_enable(void *arg, pckbport_slot_t slot, int on)
{
	/* nothing */
	return;
}

static void
vcikbc_intr_establish(void *arg, pckbport_slot_t slot)
{
	struct vcikbc_softc *sc = (struct vcikbc_softc *)arg;
	sc->sc_polling = 0;
}

static void
vcikbc_set_poll(void *arg, pckbport_slot_t slot, int on)
{
	struct vcikbc_softc *sc = (struct vcikbc_softc *)arg;
	if (on)
		sc->sc_polling = 1;
	else
		sc->sc_polling = 0;
}
